/* Demonstration program for matrix operations involved in
   LU factorization */

#include <stdio.h>
#include <string.h>

/* Maximum size for demonstration */
#define N 10

int A[N][N], G[N][N];	  /* Represent coefficient matrix, type */
int Df[N][N], Du[N][N];	  /* Down Conveyor, flag, coefficient */
int R[N][N];		  /* Right Conveyor */
int S[N][N];		  /* Schedule */
int M[N][N];		  /* Marks cells to be flagged */

/* Entries:
  A[i,j]  = k >= 0  value is A[i,j](k), -1 - L[i,j], -2 - P[i,i], -3 - U[i,j]
  G[i,j]  = 0 - below, 1 - on diag, 2 - above
  Df[i,j] = 0 - L, 1 - P, 2 - U
  Du[i,j] = row from which entry originated, or -1
  R[i,j]  = k if entry is L[i,k], -1 if nothing
  S[i,j]  = cycle on which entry should be inserted to conveyor, -1 if none.

/* Make matrices of size n by n, doesn't include S */

void start(n)
 int n;
 {int i, j;
  for ( i = 0; i < n; i++ )
  {
   for ( j = 0; j < n; j++ )
    { G[i][j] = (j >= i) + (j > i);
      A[i][j] = (i == 0) ? (-1 - G[i][j]) : 0; 
      Df[i][j] = 0;  Du[i][j] = -1;
      R[i][j] = -1;  M[i][j] = 0;
       } };
 }

int full;    /* Display format, full if true, compact if false */

/* Draw horizontal line for n cells */

horiz(n)
 int n;
 {int k;  char sg[10];
  strcpy(sg, (full) ? "+------" : "+---");
  for ( k = 0; k < n; k++ ) printf(sg);
  printf("+\n");
 }

/* Display item i, j of type g index k, g = -1 means original entry.
   Flag if m and g >= 0 */

ditem(i, j, g, k, m)
 int i, j, g, k, m;
 {if (k < 0)  printf("| -- "); else
   switch (g)
  {case -1: printf("|A%d%d(%d)", i, j, k);  break;
   case  0: printf("|L%d%d ", i, j);  break;
   case  1: printf("|P%d%d ", i, j);  break;
   case  2: printf("|U%d%d ", i, j);  break;
   default: printf("|ERR "); break; }
   if (g >= 0) { if (m) printf("* "); else printf("  "); }
  }

/* Display current status for size n, flag according to M if m.
   Erases marks. */

display(n, m)
 int n, m;
 {int i, j, x, g, k;
  for ( i = 0; i < n; i++ )
   {horiz(n);
    if (!full)
    { for ( j = 0; j < n; j++ ) printf( (M[i][j])?"| * ":"|   " );
      printf("|\n"); }
    else
    {for ( j = 0; j < n; j++ )
      { x = A[i][j];
        if (x >= 0) { g = -1;  k = x; } else { g = -1 - x;  k = 0; }
        ditem(i, j, g, k, 0); }
     printf("|\n");
     for ( j = 0; j < n; j++ )
      { x = Du[i][j];  ditem(x, j, Df[i][j], x, m && M[i][j]); }
     printf("|\n");
     for ( j = 0; j < n; j++ )
      { x = R[i][j];  ditem(i, x, 0, x, m && M[i][j]); }
     printf("|\n"); }
   };
  horiz(n);
 }

/* Print matrix for debugging */

pmat(Mx, n)
 int Mx[N][N], n;
 { int i, j;
   for ( i = 0; i < n; i++ )
    { for ( j = 0; j < n; j++ ) printf("%d ", Mx[i][j]);
      printf("\n"); }
 }

/* Shift down 1 position */

dn(n)
 int n;
 {int i, j;
  loadm(n, 0);
  for ( i = n-1; i >= 1; i-- )
    for ( j = 0; j < n; j++ )
     { Df[i][j] = Df[i-1][j];  Du[i][j] = Du[i-1][j]; }
  for ( j = 0; j < n; j++ ) { Df[0][j] = 0;  Du[0][j] = -1; }
   }

/* Shift right 1 position */

rt(n)
 int n;
 {int i, j, r;
  loadm(n, 0);
  for ( j = n-1; j >=1; j-- )
    for ( i = 0; i < n; i++ ) R[i][j] = R[i][j-1];
  for ( i = 0; i < n; i++ ) R[i][0] = -1;
   }

/* Place one entry onto down conveyor */

edn(a, b)
 int a, b;
 { Df[a][b] = G[a][b];  Du[a][b] = a; }

/* Place one entry onto right conveyor */

ert(a, b)
 int a, b;
 { R[a][b] = b; }

/* Load all mult flags with f */

loadm(n, f)
 int n, f;
 {int i, j;
  for ( i = 0; i < n; i++ )
   for ( j = 0; j < n; j++ ) M[i][j] = f;
 }

/* Spread row r down (P's and U's) */

sdn(n, r)
 int n, r;
 {int i, j, f, x;
  for ( j = 0; j < n; j++ )
   { f = G[r][j];  x = (j < r) ? -1 : r;
     for ( i = 0;  i <= r; i++ )
       { Df[i][j] = 0;  Du[i][j] = -1;  M[i][j] = 0; }
     for ( i = r+1;  i < n; i++ )
       { Df[i][j] = f;  Du[i][j] = x;  M[i][j] = 0; }
    }
  }

/* Spread newly computed L's right */

srt(n)
 int n;
 {int i, j, x;
  for ( i = 0; i < n; i++ )
   { for ( j = 0; (!M[i][j]) && (j < n); j++ ) R[i][j] = -1;
     x = j;
     if ( x < n ) { R[i][x] = -1;  M[i][x] = 0; }
     for ( j = x+1; j < n; j++ ) R[i][j] = x;
    }
  }

/* Calculate L's wherever D has a Pivot, sets marks */

calcl(n)
 int n;
 {int i, j;
  for ( i = 1; i < n; i++ )
   for ( j = 0;  j < n; j++ )
    if (Df[i][j] == 1) { A[i][j] = -1;  M[i][j] = 1; } else M[i][j] = 0;
  }

/* Calculate new A's wherever R has an L and D has an entry, sets marks */ 

calca(n)
 int n;
 {int i, j, x, y;
  for ( i = 1; i < n; i++ )
   for ( j = 0;  j < n; j++ )
    { x = R[i][j];  y = Du[i][j];
      if ( x >= 0 && y >= 0 && x == y && Df[i][j] == 2 )
	{ x = ++A[i][j];
	  if (x == i) A[i][j] = -1 - G[i][j];
          M[i][j] = 1; }
      else M[i][j] = 0; }
  }

/* Demonstrate spreading LU-factorization */

dspread(n)
 int n;
 { int i;
   start(n);  display(n, 0);  
   for ( i = 0; i < n-1; i++ )
    { sdn(n, i);  calcl(n);  pause();  display(n, 1);
      srt(n);     calca(n);  pause();  display(n, 1); }
  }

/* Demonstrate row-systolic, column spreading LU-factorization */

drowsys(n)
 int n;
 {int i;
   start(n);  display(n, 0);  
   for ( i = 0; i < n-2; i++ )
    { injctr(n, i);
      dn(n);   calcl(n);  pause();  display(n, 1);
      srt(n);  calca(n);  pause();  display(n, 1);
      dn(n);   calcl(n);  pause();  display(n, 1);
      srt(n);  calca(n);  pause();  display(n, 1); }
   injctr(n, i);
   dn(n);  calcl(n);  pause();  display(n, 1);
   srt(n);  calca(n);  pause();  display(n, 1);
  }

/* Calculate schedule */

sched(n)
 int n;
 {int i, j;
  for ( i = 0; i < n; i++ )
   for ( j = 0; j < n; j++ )
    S[i][j] = ( j <= i) ? (j+j+i) : (i+i+j);
  /* pmat(S, n); */
 }

/* Insert L's, P's, and U's according to schedule */

isched(n, c)
 int n, c;
 {int i, j, g;
  for ( i = 0; i < n; i++ )
   for ( j = 0; j < n; j++ )
    { if (S[i][j] == c)
       { g = G[i][j];
         if (g == 0) ert(i, j); else edn(i, j); }}
 }

/* Demonstrate fully systolic LU-factorization */

dsystol(n)
 int n;
 {int c, stop;
  start(n);  display(n, 0);  sched(n);  stop = 3*(n-2) + 1;
    for ( c = 0; c <= stop; c++ )
     { isched(n, c);  dn(n);  rt(n);
       calcl(n);  pause();  display(n, 1);  
       calca(n);  pause();  display(n, 1); }
 }  

/* Inject row r in down conveyor */

injctr(n, r)
 int n, r;
 {int j;
  loadm(n, 0);
  for ( j = 0; j < r; j++ ) { Df[r][j] = 0;  Du[r][j] = -1; }
  for ( j = r; j < n; j++ ) { Df[r][j] = G[r][j];  Du[r][j] = r; }
  }

/* Pause for signal from user */

pause()
 {char c[2];
  printf("Type any character to proceed>");  scanf("%1s", c);
 } 

/* Test program.  Commands are
   m <n>  - initialize matrix of size n by n
   p      - print current status
   c      - change display format (toggles between full and compact)
   o      - return to original status
   D <k>  - demonstration k on current size
	    1 - spreading
	    2 - row systolic, spreading across columns
	    3 - fully systolic
   Sd <r> - spread down - row r, entries [r...n) 
   Sr     - spread right - newly computed L's
   I <r>  - inject row r into down conveyor
   L      - calculate L's
   A      - calculate new A's
   r d    - shift {right down}
   ed <a> <b> - Place entry [a,b] on down conveyor
   er <a> <b> - Place entry [a,b] on right conveyor
   q      - quit
*/

void main()
 { char c[2];  int n, a, b;
   n = 4;  full = 1;
   while ( printf("matrix>"), scanf("%1s", c), c[0] != 'q' )
    switch (c[0])
     { case 'm':  scanf("%d", &n);
       case 'o':  if (n <= 2 || n > N) printf("Bad n\n");
                  else { start(n); /* display(n, 0); */ };
                  break;
       case 'c':  full = !full;  break;
       case 'D':
	switch ( scanf("%d", &a), a )
	 { case  1:  dspread(n);  break;
	   case  2:  drowsys(n);  break;
	   case  3:  dsystol(n);  break;
	   default:  printf("Bad input - ignored\n");  break; }
	break;
       case 'S':
        switch ( scanf("%1s", c), c[0] )
	 { case 'd':  scanf("%d", &a);
	              sdn(n, a);  break;
	   case 'r':  srt(n);  break;
	   default:   printf("Bad input - ignored\n");  break; }
	break;
       case 'I':  scanf("%d", &a);  injctr(n, a);  break;
       case 'L':  calcl(n);  break;
       case 'A':  calca(n);  break;
       case 'p':  display(n, 1);  break;
       case 'r':  rt(n);  break; 
       case 'd':  dn(n);  break;
       case 'e':  scanf("%1s%d%d", c, &a, &b);
	switch ( c[0] )
	 { case 'd':  edn(a, b);  break;
	   case 'r':  ert(a, b);  break;
	   default:   printf("Bad input - ignored\n");  break; }
	break;
       default:   printf("Bad input - ignored\n");  break;
     };
 }
